package dmc;

/*
 * TabbedChat.java
 * Matthew Wronka <wronkm@rpi.edu>
 *
 * Tabbed dialog for all active chats on a connection.
 * Also includes callbacks which implement RMsgHandler for use
 * with conversations on this connection.
 *
 * Copyright (C) 2002 Matthew Wronka <wronkm@rpi.edu>
 * This file is released under the GNU General Public License version 2,
 * a version of which should have been provided along with this file.
 *
 * You might also find a copy at:
 * http://dmc.sf.net/files/gpl.txt
 * http://www.fsf.org/copyleft/gpl.html
 */


import javax.swing.*;
import javax.swing.text.*;
import javax.swing.tree.*;
import javax.swing.event.*;

import java.awt.*;
import java.awt.event.*;

import java.net.*;
import java.io.IOException;

import wtalk.WTalk;
import wtalk.RMsgHandler;


/**
 * Handles all (group) chat messages.
 *
 * @author Matthew Wronka <wronkm@rpi.edu>
 * @version 0
 */
class ChatReader implements RMsgHandler {
	protected TabbedChat tc;

	/**
	 * Default ctor.
	 *
	 */
	public ChatReader(){}

        /**
         * Sets the parent TabbedChat object.   
         *
         * @param t Parent TabbedChat object.
         */
	public void setTabs(TabbedChat t){
		tc = t;
	}

	/**
	 * Implements handle to handle received messages.
	 *
	 * @param to The message target.
	 * @param from The message sender.
	 * @param data The message contents.
	 */
	public void handle(String to, String from, String data) {
		Document doc;
		JTextPane tp;
		int i = tc.getUI().indexOfTab(to);
		tp=tc.openBuddy(to);

		// Highlight
		i=tc.getUI().indexOfTab(to);
		if(i!=tc.getUI().getSelectedIndex()){
			tc.getUI().setForegroundAt(i, new Color(100, 0, 0));
		}

		// Highlight the connection tab
		if(tc.getTabbedConnection() != null) {
			i=tc.getTabbedConnection().getUI().indexOfTab(tc.getConnectionName());
		        if(i!=tc.getTabbedConnection().getUI().getSelectedIndex()){
                	        tc.getTabbedConnection().getUI().setForegroundAt(i, new Color(100, 0, 0));
	                }	
		}


		doc= tp.getDocument();

		try {
			doc.insertString(doc.getLength(), from + ": " + data + "\n", tp.getStyle("chat"));
			tp.scrollRectToVisible(new Rectangle(0, tp.getHeight(), 1, 1));
		} catch (BadLocationException ble) {}
	}
}

/**
 * Handles all brodcase (wall) messages.
 *
 * @author Matthew Wronka <wronkm@rpi.edu>
 * @version 0
 */
class WallReader implements RMsgHandler {
	private Document doc;
	private JTextPane tp;

        /**
         * Default ctor.
         *
         */
	public WallReader(){}

	/**
	 * Overloaded ctor to set JTextPane to use.
	 *
	 * @param t JTextPane to use when displaying walls.
	 */
	public WallReader(JTextPane t){
		setPane(t);
	}

        /**
         * Sets the parent TabbedChat object.   
         *
         * @param t Parent TabbedChat object.
         */
	public void setPane(JTextPane t){
		tp = t; doc = t.getDocument();
	}

        /**
         * Implements handle to handle received messages.
         *
         * @param to The message target. This may be null.
         * @param from The message sender.
         * @param data The message contents.
         */
	public void handle(String to, String from, String data) {
		try {
			doc.insertString(doc.getLength(), from + ": " + data + "\n", tp.getStyle("wall"));
			tp.scrollRectToVisible(new Rectangle(0, tp.getHeight(), 1, 1));
		} catch (BadLocationException ble) {}
	}
}

/**
 * Handles all system messages.
 *
 * @author Matthew Wronka <wronkm@rpi.edu>
 * @version 0
 */
class SystemReader implements RMsgHandler {
	private Document doc;
	private JTextPane tp;

        /**
         * Default ctor.
         *
         */
	public SystemReader(){}

        /**
         * Overloaded ctor to set JTextPane to use.
         *
         * @param t JTextPane to use when displaying walls.
         */
	public SystemReader(JTextPane t){
		setPane(t);
	}

        /**
         * Sets the parent TabbedChat object.   
         *
         * @param t Parent TabbedChat object.
         */
	public void setPane(JTextPane t){
		tp = t; doc = t.getDocument();
	}

        /**
         * Implements handle to handle received messages.
         *
         * @param to The message target. This may be null.
         * @param from The message sender. This may be null.
         * @param data The message contents.
         */
	public void handle(String to, String from, String data) {
		try {
			doc.insertString(doc.getLength(), "--" + data + "\n", tp.getStyle("server"));
			tp.scrollRectToVisible(new Rectangle(0, tp.getHeight(), 1, 1));
		} catch (BadLocationException ble) {}
	}
}

/**
 * Handles all private (person to person)  messages.
 *
 * @author Matthew Wronka <wronkm@rpi.edu>
 * @version 0
 */
class MsgReader implements RMsgHandler {
	protected TabbedChat tc;

        /**
         * Default ctor.
         *
         */
	public MsgReader(){}

        /**
         * Sets the parent TabbedChat object.   
         *
         * @param t Parent TabbedChat object.
         */
	public void setTabs(TabbedChat t){ 
		tc = t;
	}

        /**
         * Implements handle to handle received messages.
	 * This class also supports client->client encryption of message data.
         *
         * @param to The message target.
         * @param from The message sender.  This may be null when this class
	 *	is being used to echo outgoig messages, such as in the case
	 *	where the user does not receive messages sent.
         * @param data The message contents.
         */
	public void handle(String to, String from, String data){
		Document doc;
		JTextPane tp;
		String ciphertext=data;
		String keytype = null;
		String cipher=null;
		java.util.StringTokenizer st = new java.util.StringTokenizer(data);
		boolean isReceiver = (from != null);	//(from != tc.getWTalkProtocol().getName()); 	// Receiver or Sender?
		String s = isReceiver? from : to;
		int i = tc.getUI().indexOfTab(s);

		if(i>=0)
			tp = ((JChatPanel)(tc.getUI().getComponentAt(i))).getJTextPane();
		else
			tp = tc.addTab(s, JChatPanel.IM_TAB).getJTextPane();

		i = tc.getUI().indexOfTab(s);
		ChatInputHandler ci = ((JChatPanel)(tc.getUI().getComponentAt(i))).getChatInputHandler();


		// They want us to send our public key
		if(data.charAt(0)==1){
			System.out.println("1...");
			ci.sendPublicKey();
			return;
		}


		if(data.charAt(0)==2){
			System.out.println("2...");
			// Did we get a key?
			if(st.hasMoreTokens() ) keytype = st.nextToken();
			//if( keytype.equals("dmc-key")) {
				if(isReceiver && st.hasMoreTokens()){
					ci.setPublicKey(data.substring(keytype.length()+1));
					ci.negotiateSecure();
				}
				//return;
			//}
			return;
		}


		if(data.charAt(0)==3) {
			System.out.println("3...");
			ciphertext =ci.decryptRSA(data.substring(1));
			//st = new java.util.StringTokenizer(ciphertext);
			//if(st.hasMoreTokens() ) cipher= st.nextToken();
			// changing encryption protocols?
			if( isReceiver ){
				ci.manageSecure("blowfish");
				ci.setSharedKey(ciphertext.getBytes());
				//ci.manageSecure(cipher);
				//System.out.println("New cipher: " + cipher);
				//if(ci.manageSecure("blowfish")){
					//if(st.hasMoreTokens())
						//ci.setSharedKey(st.nextToken().getBytes());
					return;
				//}
			}
			return;
		}

		ciphertext = data;		
		data = ci.decrypt(ciphertext);
		if (data==null) 
			if(isReceiver)
				data = cipher;
			else return;
		

		// Highlight
		i=tc.getUI().indexOfTab(s);
		if(i!=tc.getUI().getSelectedIndex()){
			tc.getUI().setForegroundAt(i, DmcColors.TAB_HIGHLIGHT);
		}


                // Highlight connection ....
		if(tc.getTabbedConnection() != null) {
			i=tc.getTabbedConnection().getUI().indexOfTab(
				tc.getConnectionName());
       			if(i!=tc.getTabbedConnection().getUI().getSelectedIndex())
				 tc.getTabbedConnection().getUI().setForegroundAt(i, DmcColors.TAB_HIGHLIGHT);
		}
     
		

		doc = tp.getDocument();

		try {
			if(isReceiver)
				doc.insertString(doc.getLength(), from + ": ", tp.getStyle("usay"));
			else
				doc.insertString(doc.getLength(), tc.getUserName() + ": ", tp.getStyle("isay"));
			doc.insertString(doc.getLength(), data + "\n", tp.getStyle("chat"));
			tp.scrollRectToVisible(new Rectangle(0, tp.getHeight(), 1, 1));
		} catch (BadLocationException ble) {}
		
	}
}

/**
 * Handles all automated messages.
 *
 * @author Matthew Wronka <wronkm@rpi.edu>
 * @version 2002-11-26
 */
class AutoReader implements  RMsgHandler {
	TabbedChat tc=null;

        /**
         * Default ctor.
         *
         */
	public AutoReader(){}

	/**
	 * Sets the parent TabbedChat object.
	 *
	 * @param t Parent TabbedChat object.
	 */
	public void setTabs(TabbedChat t){ tc = t;}

        /**
         * Implements handle to handle received messages.
         *
         * @param to The message target.
         * @param from The message sender.
         * @param data The message contents.
         */
	public void handle(String to, String from, String data) {
		JTextPane tp = tc.openBuddy(from);
		Document doc = tp.getDocument();
		int i=tc.getUI().indexOfTab(from);

		if(i!=tc.getUI().getSelectedIndex()){
			tc.getUI().setForegroundAt(i, DmcColors.TAB_HIGHLIGHT);
		}


		try {
			doc.insertString(doc.getLength(), data+"\n", tp.getStyle("auto"));
			tp.scrollRectToVisible(new Rectangle(0, tp.getHeight(), 1, 1));
		}catch (BadLocationException ble){}

	}

	
}


/**
 * Manages chats with multiple targets in a single JTabbedPane.
 * Also provides several functions for accessing information on 
 * the active connection.
 *
 * @author Matthew Wronka <wronkm@rpi.edu>
 * @version 0
 */
public class TabbedChat  {
	final protected JTabbedPane tabbedPane = new JTabbedPane();
	final protected ChatReader cr = new ChatReader();
	final protected WallReader wr = new WallReader();
	final protected SystemReader sr = new SystemReader();
	final protected MsgReader mr = new MsgReader();
	final protected AutoReader ar = new AutoReader();
	protected WTalk wt = null;
	protected wtalk.IUserList ul = null;
	protected TabbedConnection m_tc = null;
	protected String m_ConnectionName = null;


	/**
	 * Default ctor. 
	 */
	public TabbedChat() {
		InitTabbedChat();
		//setConnectionType(wtalk.Protocol.USA);
		//setConnectionType(wtalk.Protocol.PMS);
		//setConnectionType(wtalk.Protocol.POP);
	}

	/**
	 * Overloaded ctor.
	 *
	 * @param tc Parent TabbedConnection containing this TabbedChat.
	 * @param name Connection's name.
	 */
	public TabbedChat(TabbedConnection tc, String name) {
		m_tc = tc;
		m_ConnectionName = name;
		wt = new WTalk(m_ConnectionName);
		InitTabbedChat();
	}

	/**
	 * Initilizes TabbedChat
	 */
	public void InitTabbedChat() {
		tabbedPane.setFocusable(false);
		initReaders();
		addListeners();
		addKeyBindings();
	}

	/**
	 * Returns the parent TabbedConnection.
	 *
	 * @return  TabbedConnection containing this TabbedChat.
	 */
	public TabbedConnection getTabbedConnection(){ return m_tc; }

	/**
	 * Returns the current connection name.
	 *
	 * @return Returns the name of the current connection.
	 */
	public String getConnectionName(){ return m_ConnectionName;}

	/**
	 * Sets the protocol used to connect.  This should be a constant
	 * defined in wtalk.Protocol.
	 *
	 * @param i Protocol to use, as defined in wtalk.Protocol.
	 * @see wtalk.WTalk
	 * @see wtalk.Protocol
	 */
	public void setConnectionType(int i) {
		wt.setWTalkProtocol(i);
		registerReaders();
	}

	/**
	 * Initializes message readers.
	 */
	private void initReaders() {
		JChatPanel cp = addTab("System", JChatPanel.SYSTEM_TAB);
		wr.setPane(cp.getJTextPane());
		sr.setPane(cp.getJTextPane());
		cr.setTabs(this);
		mr.setTabs(this);
		ar.setTabs(this);
	}

	/**
	 * Register readers with wtalk.WTalk object.
	 */
	private  void registerReaders() {
		if(wt.getWTalkProtocol()==null) {
			System.err.println("WTalk Protocol is UNDEF!");
			return;
		}

		wt.getWTalkProtocol().setReceivedMessageHandler(cr);
		wt.getWTalkProtocol().setChatMessageHandler(cr);
		wt.getWTalkProtocol().setPrivateMessageHandler(mr);
		wt.getWTalkProtocol().setWallMessageHandler(wr);
		wt.getWTalkProtocol().setSystemMessageHandler(sr);
		wt.getWTalkProtocol().setAutoMessageHandler(ar);
		wt.getWTalkProtocol().setUserListHandler(ul);
	}

	/**
	 * Sets the object to register with WTalkProtocols that will
	 * implement IUserList to support GUI representation of buddy lists.
	 *
	 * @param u Object that will be used to manage user lists.
	 */
	public void setUserList(wtalk.IUserList u) { 
		ul = u; 
		wt.getWTalkProtocol().setUserListHandler(ul);
	}

	/**
	 * Adds event listeners to tabs to facilitate highlighting of 
	 * tabs which do not have focus when they receive messages.
	 */
	private void addListeners() {
		tabbedPane.addChangeListener(new ChangeListener(){
			public void stateChanged(ChangeEvent e) {
				tabbedPane.setForegroundAt(tabbedPane.getSelectedIndex(), DmcColors.TAB_NORMAL);
			}
		});
	}

	 /**
	 * Adds keybindigns to facilitate switching conversations.
	 */
	private void addKeyBindings() {
		tabbedPane.getInputMap(
			JComponent.WHEN_IN_FOCUSED_WINDOW).
			 put(KeyStroke.getKeyStroke("ctrl H"), "next-conversation") ;
		tabbedPane.getInputMap(
			 JComponent.WHEN_IN_FOCUSED_WINDOW).
			put(KeyStroke.getKeyStroke("ctrl L"), "prev-conversation") ;

		Action PrevConversation = new AbstractAction(){
			public void actionPerformed(ActionEvent e){
				int i=tabbedPane.getSelectedIndex();
				if(i==tabbedPane.getTabCount()-1)
					tabbedPane.setSelectedIndex(0);
				else
					tabbedPane.setSelectedIndex(i+1);
			}
		 };

		Action NextConversation = new AbstractAction(){
			public void actionPerformed(ActionEvent e){
				int i=tabbedPane.getSelectedIndex();
				if(i==0)
					tabbedPane.setSelectedIndex(
						tabbedPane.getTabCount()-1);
				else
					tabbedPane.setSelectedIndex(i-1);

			}
		};


		tabbedPane.getActionMap().put("next-conversation", NextConversation) ;
		tabbedPane.getActionMap().put("prev-conversation", PrevConversation) ;
	}


	/**
	 * Returns the GUI representation of a TabbedChat.
	 *
	 * @return Returns GUI  representation of this object.
	 */
	public JTabbedPane getUI() {
		return tabbedPane;
	}

	/**
	 * Returns a ChatInputHandler associated with the JChatPanel
	 * in charge of <em>room</em>.
	 *
	 * @param room	Room for which we want to get a ChatInputHandler.
	 * @return ChatInputHandler for <em>room</em>
	 */
	public ChatInputHandler getInputPanel(String room) {
		int i = getUI().indexOfTab(room); 
		return ((JChatPanel)(getUI().getComponentAt(i))).getChatInputHandler();
	}

	/**
	 * Returns  the user name as reported by the WTalkProtocol.
	 *
	 * @return User name assocaited with the active connection.
	 */
	public String getUserName() {
		return wt.getWTalkProtocol().getName();
	}

	/**
	 * Tells the WTalkProtocol to  request a new name, <em>name</em>.
	 * There is no guarantee that this name will be granted
	 *
	 * @param name Name to request.
	 */
	public void setName(String name) {
		wt.getWTalkProtocol().setName(name);
	}

	/**
	 * Tells the WTalkProtocol to use the given password when trying
	 * to authenticate.
	 *
	 * @param passwd Password to use to authenticate.
	 */
	public void setPassword(String passwd) {
		wt.getWTalkProtocol().setPassword(passwd);
	}


	/**
	 * Connects to <em>server:port</em> using <em>login</em> as a handle.
	 *
	 * @param server Host to connect to.
	 * @param port Port to connect to on host.
	 * @param login Name to login with.
	 * @throws IOException Throws IOException when the WTalkProtocol cannot
	 *	connect.  For instance, the host is down, or cannot be resolved.
	 */
	public void connect(String server, int port, String login) throws IOException{
		wt.getWTalkProtocol().connect(server, port, login);
	}

	/**
	 * Adds a new tab to the TabbedChat with caption <em>name</em>.
	 *	
 	 * @param name Caption for tab. 
	 * @param type Type of JChatPanel to create.  
	 *	These are defined in JChatPanel.
	 * @return JChatPanel that belongs to the new tab.
	 * @see JChatPanel
	 */
	public JChatPanel addTab(String name, int type) {
		JChatPanel pane = new JChatPanel(this, wt, type!=0?name:null);
		
		pane.setBorder(BorderFactory.createEmptyBorder(10, 10, 10, 10));

		tabbedPane.addTab(name, pane);
		tabbedPane.setForegroundAt(tabbedPane.indexOfTab(name), DmcColors.TAB_NORMAL);

		return pane;
	}

	/**
	 * Returns the JTextPane used by the JChatPanel displaying the 
	 *	chat with <em>name</em>.  If none exists, a new tab is added
	 *	which contains a JChatPanel to use in communications with
	 *	this buddy.
	 *
	 * @param name The handle of the user whose JChatPanel's JTextPane
	 *	is to be returned.
	 * @return JTextPane used by the JChatPanel that represents the chat
	 *	 with <em>name</em>.
	 */
	public JTextPane openBuddy(String name) {
		int i = getUI().indexOfTab(name);
		JTextPane tp=null;
		if(i>=0) 
			tp = ((JChatPanel)(getUI().getComponentAt(i))).getJTextPane(); 
		else
			tp = addTab(name, JChatPanel.CHAT_TAB).getJTextPane();

		return tp;
	}

}
